/*
* (c)2017 The University of Aizu
* This software is released under the New BSD License.
*/
#ifndef LOCAL_OPENGL_H
#define LOCAL_OPENGL_H

#include <time.h>
#include <stdio.h>

#include <GL/glut.h>
#include <GL/freeglut.h>

#include "local_basic_type.h"
#include "Render/Mouse.h"

/*=================================================================================================
  Macro definition
 ================================================================================================*/
#define SUB_THREAD_NG        (0)
#define OPENGL_SLEEP         (250)
#define OPENGL_CLOSE         (50)
#define THREAD_CLOSE         (1000)

#define VERTEX_NUM_MAX       (3)
#define LENGTH_MAX           (1024)
#define LENGTH_MID           (512)

#define WINDOW_SIZE_W        (72)
#define WINDOW_SIZE_H        (54)

#define POS_INIT_X           (100)
#define POS_INIT_Y           (100)

#define OPNGL_WINDOW_W       (1280)
#define OPNGL_WINDOW_H       (720)

#define WINDOW_SIZE_MIN      (1)

#define WORK_DATA_INIT       (0)
#define DEPTH_DATA_INIT      (0.0f)

#define Z_DATA_INIT          (0.0f)
#define V_DATA_INIT          (1.0f)
#define U_DATA_INIT          (1.0f)

#define RENDER_SIZE_INIT     (1.0f)
#define RENDER_SIZE_MAX      (4.0f)

#define DEF_WIRE_FRAME_ON    (1)
#define DEF_WIRE_FRAME_OFF   (0)

#define DEFINE_FRAME_ON      ("wire_frame_on")
#define DEFINE_FRAME_OFF     ("wire_frame_of")

#define Z_COORD_INIT         RENDER_SIZE_INIT
#define Z_COORD_MIN          (1.0f)
#define Z_COORD_MAX          (30.0f)
#define Z_COORD_VAL          (0.9f)

#define DEF_PLAY_MODE        (1)
#define DEF_STOP_MODE        (0)

#define DEFINE_MODE_PLY      ("play")
#define DEFINE_MODE_STP      ("stop")

#define DF_EMPTY_STRING      ("")

#define READ_FILE_SUCCESS    ("Successful reading of the 3D image object file.")

#define EXEC_STR             (0)
#define EXEC_END             (1)
#define EXEC_INIT            EXEC_END
#define GUARD_TIM_INI        (0)
#define GUARD_TIM_MAX        (100)

/*=================================================================================================
  variable type specification
 ================================================================================================*/
  struct Vertex{
    union{
      struct{
        float x;
        float y;
        float z;
      };
      float v[VERTEX_NUM_MAX];
    };
    operator float* () { return (float*)&x; }
    operator const float* () { return (float*)&x; }
    Vertex &operator = (const Vertex &vert) { x=vert.x; y=vert.y; z=vert.z; return (*this); }
    Vertex operator *(float s) { return Vertex(x*s, y*s, z*s); }
    Vertex operator /(float s) { float sInv = 1.0f/s; return Vertex(x*sInv, y*sInv, z*sInv); }
    Vertex(float nx=0.0, float ny=0.0, float nz=0.0) : x(nx), y(ny), z(nz) {} 
  };

  struct UV{
    struct{
      float u;
      float v;
    };
    operator float* () { return (float*)&u; }
    operator const float* () { return (float*)&u; }
    bool operator ==(const UV &uv )const{
      if( u == uv.u && uv.v == v ) return true;
      else return false;
    }
    bool operator <(const UV& uv)const{
      if( uv.u < u && uv.v < v ) return true;
      else return false;
    }
    UV &operator = (const UV &uv ){ u = uv.u; v = uv.v; return (*this); }
    UV( float nu =0.0, float nv =0.0 ):u(nu), v(nv){}
  };

/*=================================================================================================
  prototypes definition
 ================================================================================================*/
  static void nvdGlutDisplay( void );
  static void nvdGlutUpdate( void );
  static void nvdReshape( int, int );
  static void nvdGlutInit( void );
  void glutInitialize( void );

	static unsigned int __stdcall StartGLUT(void *dummy);
  bool_type bThreadStart( void );
  void nvd3DShapeRenderObject( void );
  void ReleaseRender( void );
  void SetTexture( void );
  void nvdMakeRenderObject( void );
  void nvdRotationRender( void );

  void Mouse(int button, int state, int x, int y);
  void Motion(int x, int y);
  void PassiveMotion(int x, int y);

  bool_type bRead3DshapeFile( const char_type * );
  void nvdObjectDataInitialize( void );
  void nvdCheckOpenGLRequest( void );
  void nvdOpenGLFileCopy( void );
  void nvdOpenGLChangeWireFrame( void );
  void nvdOpenGLChangeZcoordinate( void );
  void nvdOpenGLChangePlayMode( void );

/*=================================================================================================
  variables definition
 ================================================================================================*/
	Vertex **m_vertex;
	UV     **m_uv;

  // OpenGL(freeglut2-8-1) related
  HANDLE hHandle;
  GLuint _displayList;
  GLuint _texname;

  int g_windowWidth;  // 1280
  int g_windowHeight; // 720
  int g_windowPosX;   // 100
  int g_windowPosY;   // 100

  float_type g_RGB_Red;
  float_type g_RGB_Green;
  float_type g_RGB_Blue;

  char_type src_path_ftv[LENGTH_MAX];
  char_type src_path_bmp[LENGTH_MAX];
  char_type dst_path_ftv[LENGTH_MAX];
  char_type dst_path_bmp[LENGTH_MAX];

  bool_type bReqFileChange;
  bool_type bVaidDiapFlag;

  bool_type  bReqWireFrame;
  bool_type  bWireFrameMode;
  uword_type uwWireFrameMode;

  bool_type  bReqZcoordAdjust;
	float_type _render_size;
  float_type g_fZcoordValue;

  bool_type  bReqPlayMode;
  bool_type  bPlayMode_sts;
  bool_type  bPlayStopMode;

  bool_type  bEXEC_Thread;

  ViewCamera g_ViewportCamera;

#endif // LOCAL_OPENGL_H
