// -*- C++ -*-
/*!
 * @file  rtc_3DImageShapeViewer.h
 * @brief Processing of three dimensional image Viewer
 * @date  $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#ifndef RTC_3DIMAGESHAPEVIEWER_H
#define RTC_3DIMAGESHAPEVIEWER_H

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/idl/InterfaceDataTypesSkel.h>

// Service implementation headers
#include <iostream>
#include <string>
#include <stdio.h>
#include <direct.h>
#include <Shlwapi.h>
#include <windows.h>

// add to code for OpenCV
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

// add to code
// OpenGL(freeglut2-8-1) related
#include <GL/glut.h>
#include <GL/freeglut.h>
#include <process.h>
#include <vector>

// add to code
#include "local_basic_type.h"
#include "Render/Mouse.h"

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="port_stub_h">
// </rtc-template>

using namespace RTC;

#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

// add to code
/*=================================================================================================
  Macro definition
 ================================================================================================*/
//#define _DEBUG_SW_

//#define RESULT_OK                    (1)
//#define RESULT_NG                    (0)

#define _ON_INITIALIZE                 ("onInitialize execution")
#define _ON_ACTIVATED                  ("onActivated execution")
#define _ON_DEACTIVATED                ("onDeactivated execution")
#define _ON_ABORTING                   ("onAborting execution")
#define _ON_RESET                      ("onReset execution")
#define _CREATE_WINDOW                 ("create window")
#define _DELETE_WINDOW                 ("destroy all windows")
#define _DELETE_OPENGL_WINDOW          ("close the opengl window")
#define _DELETE_OPENGL_THREAD          ("close the opengl thread")
#define _ERROR_RESET                   ("ERROR: Error was detected. Please reset the 3D image shape Viewer RTC by the RT System Editor.")

#define _ERROR_IMAGE_EMPTY             ("ERROR: Empty error of image data")
#define _ERROR_IMAGE_SIZE              ("ERROR: This image data is outside the size target")
#define _ERROR_PATH_WRONG              ("ERROR: The input directory path wrong")
#define _ERROR_INPUT_DATA              ("ERROR: Input port data wrong")

#define DEF_MAX_LENGTH                 (1024)
#define DEF_MID_LENGTH                 (512)
#define DEF_MIN_LENGTH                 (64)

#define DEF_GET_PATH_NG                (0)
#define DEF_GET_PATH_OK                (1)

#define DEF_RGB_GL0F                   (0.0)
#define DEF_RGB_GL1F                   (1.0)
#define DEF_RGB_255F                   (255.0)

/*!
 * @class rtc_3DImageShapeViewer
 * @brief Processing of three dimensional image Viewer
 *
 * Processing of three dimensional image Viewer
 *
 * 3D image object path information
 *
 * -
 *
 * -
 *
 */
class rtc_3DImageShapeViewer
  : public RTC::DataFlowComponentBase
{
 public:
  /*!
   * @brief constructor
   * @param manager Maneger Object
   */
  rtc_3DImageShapeViewer(RTC::Manager* manager);

  /*!
   * @brief destructor
   */
  ~rtc_3DImageShapeViewer();

  // <rtc-template block="public_attribute">
  
  // </rtc-template>

  // <rtc-template block="public_operation">
  
  // </rtc-template>

  /***
   * Initialization of the 3D Image Shape Viewer RTC
   *
   * The initialize action (on CREATED->ALIVE transition)
   * formaer rtc_init_entry() 
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post ｰ
   * 
   */
   virtual RTC::ReturnCode_t onInitialize();

  /***
   * 
   *
   * The finalize action (on ALIVE->END transition)
   * formaer rtc_exiting_entry()
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onFinalize();

  /***
   *
   * The startup action when ExecutionContext startup
   * former rtc_starting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);

  /***
   *
   * The shutdown action when ExecutionContext stop
   * former rtc_stopping_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);

  /***
   * Activate the 3D Image Shape Viewer RTC
   *
   * The activated action (Active state entry action)
   * former rtc_active_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ec_id);

  /***
   * Deactivate the 3D Image Shape Viewer RTC
   *
   * The deactivated action (Active state exit action)
   * former rtc_active_exit()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ec_id);

  /***
   * Periodic task of the 3D Image Shape Viewer RTC
   *
   * The execution action that is invoked periodically
   * former rtc_active_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ec_id);

  /***
   * onAborting the 3D Image Shape Viewer RTC
   *
   * The aborting action when main logic error occurred.
   * former rtc_aborting_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);

  /***
   *
   * The error action in ERROR state
   * former rtc_error_do()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);

  /***
   * Reset initialization of the 3D Image Shape Viewer RTC
   *
   * The reset action that is invoked resetting
   * This is same but different the former rtc_init_entry()
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * @pre -
   * @post -
   * 
   */
   virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  
  /***
   *
   * The state update action that is invoked after onExecute() action
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);

  /***
   *
   * The action that is invoked when execution context's rate is changed
   * no corresponding operation exists in OpenRTm-aist-0.2.0
   *
   * @param ec_id target ExecutionContext Id
   *
   * @return RTC::ReturnCode_t
   * 
   * 
   */
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


 protected:
  // <rtc-template block="protected_attribute">
  
  // </rtc-template>

  // <rtc-template block="protected_operation">
  
  // </rtc-template>

  // Configuration variable declaration
  // <rtc-template block="config_declare">
  /*!
   * Setting the background color of the viewer
   * - Name: Back_Color_Red Back_Color_Red
   * - DefaultValue: 0
   * - Range: 0<=x<=255
   */
  int m_Back_Color_Red;
  /*!
   * Setting the background color of the viewer
   * - Name: Back_Color_Green Back_Color_Green
   * - DefaultValue: 0
   * - Range: 0<=x<=255
   */
  int m_Back_Color_Green;
  /*!
   * Setting the background color of the viewer
   * - Name: Back_Color_Blue Back_Color_Blue
   * - DefaultValue: 255
   * - Range: 0<=x<=255
   */
  int m_Back_Color_Blue;

  // </rtc-template>

  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::CameraImage m_in_img;
  /*!
   * This port inputs for the camera image type.
   * - Type: RTC::CameraImage
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::CameraImage> m_in_imgIn;
  RTC::TimedString m_file_path;
  /*!
   * This port inputs the path information of the opengl file.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::TimedString> m_file_pathIn;
  RTC::TimedString m_ply_mode;
  /*!
   * This port inputs the play mode information of the 3-D
   * images-shape-system.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::TimedString> m_ply_modeIn;
  RTC::TimedString m_z_adjust;
  /*!
   * This port inputs Z coordinate adjustment data.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::TimedString> m_z_adjustIn;
  RTC::TimedString m_wire_frame_mode;
  /*!
   * This port inputs the wire-frame mode information of the 3-D
   * images-shape-system.
   * - Type: RTC::TimedString
   * - Number: 1
   * - Semantics: -
   * - Unit: -
   * - Frequency: Caused by user actions.
   * - Operation Cycle: two cycles per second.
   */
  InPort<RTC::TimedString> m_wire_frame_modeIn;
  
  // </rtc-template>


  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">
  
  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">
  
  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">
  
  // </rtc-template>

 private:
  // <rtc-template block="private_attribute">
  
  // </rtc-template>

  // <rtc-template block="private_operation">
  
  // </rtc-template>

  // add to code
  // add on prototype of function
  const uword_type WINDOW_SIZE_W    = (320);
  const uword_type WINDOW_SIZE_H    = (240);
  const uword_type IMAGE_SIZE_W_MIN = (0);
  const uword_type IMAGE_SIZE_H_MIN = (0);
  const uword_type IMAGE_SIZE_W_MAX = (640);
  const uword_type IMAGE_SIZE_H_MAX = (480);

  void nvd3DImageViewerOnInitialize( void );
  void nvd3DImageViewerOnDeactivated( void );
  void nvd3DImageViewerOnAborting( void );
  void nvd3DImageViewerOnReset( void );

  void nvd3DImageViewerOnExecute( void );
  void nvdCheckInputDirectoryPath( void );
  void nvdCreationDir3d_shape( void );
  void nvdDeleteObjFile( void );
  void nvdCopyObjFile( char_type *, char_type * );
  void nvdRelease( void );
  void nvdCheckInputWireFrameMode( void );
  void nvdCheckInputZcoordinate( void );
  void nvdCheckInputPlayStopMode( void );

  void createAppWindow( void );
  void closeAppWindow( void );
  void readCameraImage( void );
  void showCameraImage( RTC::CameraImage );

  bool_type b3DImageViewerOnActivated( void );
  bool_type bAllocationDynamicmMem( uword_type, uword_type );

  float_type fColorCodeConversion( sword_type );

  // working memory related
  char_type g_absolute_dir[DEF_MAX_LENGTH];
  char_type g_filepath_bmp[DEF_MAX_LENGTH];
  char_type g_filepath_ftv[DEF_MAX_LENGTH];

  uword_type uwAbsolute_dir_err;

  std::string winname;
};


extern "C"
{
  DLL_EXPORT void rtc_3DImageShapeViewerInit(RTC::Manager* manager);
};

#endif // RTC_3DIMAGESHAPEVIEWER_H
