// -*- C++ -*-
/*!
 * @file  rtc_3DImageShapeViewer.cpp
 * @brief Processing of three dimensional image Viewer
 * @date $Date$
 *
 * @author EAST JAPAN ACCOUNTING CENTER CO.,LTD.
 *
 * (c)2017 The University of Aizu
 * This software is released under the New BSD License.
 *
 * $Id$
 */

#include "rtc_3DImageShapeViewer.h"
#include "local_opengl.h"

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_3dimageshapeviewer_spec[] =
  {
    "implementation_id", "rtc_3DImageShapeViewer",
    "type_name",         "rtc_3DImageShapeViewer",
    "description",       "Processing of three dimensional image Viewer",
    "version",           "1.0",
    "vendor",            "EAC",
    "category",          "Image",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.Back_Color_Red", "0",
    "conf.default.Back_Color_Green", "0",
    "conf.default.Back_Color_Blue", "255",
    // Widget
    "conf.__widget__.Back_Color_Red", "slider.1",
    "conf.__widget__.Back_Color_Green", "slider.1",
    "conf.__widget__.Back_Color_Blue", "slider.1",
    // Constraints
    "conf.__constraints__.Back_Color_Red", "0<=x<=255",
    "conf.__constraints__.Back_Color_Green", "0<=x<=255",
    "conf.__constraints__.Back_Color_Blue", "0<=x<=255",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
rtc_3DImageShapeViewer::rtc_3DImageShapeViewer(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_in_imgIn("ImgIn", m_in_img),
    m_file_pathIn("opengl_file_path", m_file_path),
    m_ply_modeIn("InPlay_3DSystem", m_ply_mode),
    m_z_adjustIn("In_Z_coordinate", m_z_adjust),
    m_wire_frame_modeIn("In_wire_frame_mode", m_wire_frame_mode)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
rtc_3DImageShapeViewer::~rtc_3DImageShapeViewer()
{
  nvdRelease();
}


/*!
 * Initialization of the 3D Image Shape Viewer RTC
 */
RTC::ReturnCode_t rtc_3DImageShapeViewer::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("ImgIn", m_in_imgIn);
  addInPort("opengl_file_path", m_file_pathIn);
  addInPort("InPlay_3DSystem", m_ply_modeIn);
  addInPort("In_Z_coordinate", m_z_adjustIn);
  addInPort("In_wire_frame_mode", m_wire_frame_modeIn);
  
  // Set OutPort buffer
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Back_Color_Red", m_Back_Color_Red, "0");
  bindParameter("Back_Color_Green", m_Back_Color_Green, "0");
  bindParameter("Back_Color_Blue", m_Back_Color_Blue, "255");
  // </rtc-template>
  
  // add to code
  nvd3DImageViewerOnInitialize();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DImageShapeViewer::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DImageShapeViewer::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DImageShapeViewer::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Activate the 3D Image Shape Viewer RTC
 */

RTC::ReturnCode_t rtc_3DImageShapeViewer::onActivated(RTC::UniqueId ec_id)
{
  // add to code
  if( b3DImageViewerOnActivated() )
  {
    return RTC::RTC_OK;
  }
  else
  {
    return RTC::RTC_ERROR;
  }
}

/*!
 * Deactivate the 3D Image Shape Viewer RTC
 */

RTC::ReturnCode_t rtc_3DImageShapeViewer::onDeactivated(RTC::UniqueId ec_id)
{
  // add to code
  nvd3DImageViewerOnDeactivated();

  return RTC::RTC_OK;
}

/*!
 * Periodic task of the 3D Image Shape Viewer RTC
 */

RTC::ReturnCode_t rtc_3DImageShapeViewer::onExecute(RTC::UniqueId ec_id)
{
  // add to code
  nvd3DImageViewerOnExecute();

  return RTC::RTC_OK;
}

/*!
 * onAborting the 3D Image Shape Viewer RTC
 */

RTC::ReturnCode_t rtc_3DImageShapeViewer::onAborting(RTC::UniqueId ec_id)
{
  // add to code
  nvd3DImageViewerOnAborting();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DImageShapeViewer::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * Reset initialization of the 3D Image Shape Viewer RTC
 */

RTC::ReturnCode_t rtc_3DImageShapeViewer::onReset(RTC::UniqueId ec_id)
{
  // add to code
  nvd3DImageViewerOnReset();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t rtc_3DImageShapeViewer::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t rtc_3DImageShapeViewer::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/***************************************************************************************************
 *  Function        : bThreadRun
 *  Description     : This function starts a thread for the OpenGL.
 *  Argument        : nothing special
 *  Return          : information of normal end or abnormal end
 **************************************************************************************************/
bool_type bThreadStart( void )
{
  bool_type abReturn = FALSE;

  hHandle = (HANDLE)_beginthreadex(NULL, 0, StartGLUT, NULL, 0, NULL);
  if( hHandle != SUB_THREAD_NG )
  {
    abReturn = TRUE;
  }

  return abReturn;
}

/***************************************************************************************************
 *  Function        : StartGLUT
 *  Description     : This function calls OpenGL initialization and mainloop.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
unsigned int __stdcall StartGLUT(void *dummy)
{
  std::cout << "StartGLUT" << std::endl;

  int argc = 1;
  char_type *argv[1] = {"Free viewpoint television"};

  nvdGlutInit();

  glutInit( &argc, argv );
  glutInitWindowPosition( g_windowPosX, g_windowPosY );
  glutInitWindowSize( g_windowWidth, g_windowHeight );
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(nvdGlutDisplay);
  glutReshapeFunc(nvdReshape);
  glutIdleFunc(nvdGlutUpdate);
  glutMouseFunc(Mouse);
  glutMotionFunc(Motion);
  glutPassiveMotionFunc(PassiveMotion);
  glutInitialize();

  glutInitWindowPosition( g_windowPosX, g_windowPosY );
  glutInitWindowSize( g_windowWidth, g_windowHeight );
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);

  glutSetOption( GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS );
  bEXEC_Thread = EXEC_STR;
  glutMainLoop();
  bEXEC_Thread = EXEC_END;

  _endthreadex(0);

  return TRUE;
}

/***************************************************************************************************
 *  Function        : nvdGlutInit
 *  Description     : This function execution initialization of OpenGL.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void nvdGlutInit( void )
{
  g_windowPosX   = POS_INIT_X;
  g_windowPosY   = POS_INIT_Y;
  g_windowWidth  = OPNGL_WINDOW_W;
  g_windowHeight = OPNGL_WINDOW_H;
}

/***************************************************************************************************
 *  Function        : nvdGlutDisplay
 *  Description     : This function sets the display data of OpenGL.
 *  Argument        : 
 *  Return          : 
 **************************************************************************************************/
void nvdGlutDisplay( void )
{
  if ( bVaidDiapFlag )
  {
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glViewport( 0, 0, g_windowWidth, g_windowHeight );

    if( (bPlayStopMode == DEF_PLAY_MODE) )
    {
      glMatrixMode( GL_PROJECTION );
      glLoadIdentity();
      gluPerspective( 40.0, (double)g_windowWidth/(double)g_windowHeight, 0.1, 10000.0 );
      glMatrixMode( GL_MODELVIEW );
      glLoadIdentity();

      nvd3DShapeRenderObject();

      glPushMatrix();

      g_ViewportCamera.Set();

      if( bWireFrameMode ) glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

      nvdRotationRender();

      if( bWireFrameMode ) glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

      glPopMatrix();
    }

    glutSwapBuffers();
  }

  nvdCheckOpenGLRequest();

  Sleep(OPENGL_SLEEP);
}

/***************************************************************************************************
 *  Function        : nvdReshape
 *  Description     : This function resizes the OpenGL window.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void nvdReshape(int x, int y)
{
  g_windowWidth  = x;
  g_windowHeight = y;
  if ( g_windowWidth < WINDOW_SIZE_MIN )  g_windowWidth  = WINDOW_SIZE_MIN;
  if ( g_windowHeight < WINDOW_SIZE_MIN ) g_windowHeight = WINDOW_SIZE_MIN;
}

/***************************************************************************************************
 *  Function        : nvdGlutUpdate
 *  Description     : This function performs the OpenGL redraw.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void nvdGlutUpdate( void )
{
  glutPostRedisplay();
}

/***************************************************************************************************
 *  Function        : Mouse
 *  Description     : This function inputs the state of mouse operation.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void Mouse(int button, int state, int x, int y)
{
  if( (bVaidDiapFlag == TRUE) && (bPlayStopMode == DEF_PLAY_MODE) )
  {
    g_ViewportCamera.MouseInput(button, state, x, y);
  }
}

/***************************************************************************************************
 *  Function        : Motion
 *  Description     : This function inputs mouse motion.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void Motion(int x, int y)
{
  if( (bVaidDiapFlag == TRUE) && (bPlayStopMode == DEF_PLAY_MODE) )
  {
    g_ViewportCamera.MouseMotion(x, y);
  }
}

/***************************************************************************************************
 *  Function        : PassiveMotion
 *  Description     : 
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void PassiveMotion(int x, int y)
{
}

/***************************************************************************************************
 *  Function        : glutInitialize
 *  Description     : This function sets the screen color in blue.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void glutInitialize( void )
{
  glClearColor(g_RGB_Red, g_RGB_Green, g_RGB_Blue, 1.0);
  glEnable( GL_DEPTH_TEST );
  glShadeModel( GL_SMOOTH );
}

/***************************************************************************************************
 *  Function        : nvd3DShapeRenderObject
 *  Description     : 
 *  Argument        : 
 *  Return          : 
 **************************************************************************************************/
void nvd3DShapeRenderObject( void )
{
  ReleaseRender();
  SetTexture();
  nvdMakeRenderObject();
}

/***************************************************************************************************
 *  Function        : ReleaseRender
 *  Description     : This function deletes the display list.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void ReleaseRender( void )
{
  glDeleteLists( _displayList, 1 );
  glDeleteTextures( 1, &_texname );
}

/***************************************************************************************************
 *  Function        : SetTexture
 *  Description     : This function sets the texture.
 *  Argument        : image data
 *  Return          : nothing special
 **************************************************************************************************/
void SetTexture( void )
{
  bool_type bResult = TRUE;
  cv::Mat img;

  try{
    img = cv::imread( &dst_path_bmp[0] );
  }
  catch (...){
    bResult = FALSE;
  }

  if ( bResult == TRUE )
  {
    if (img.empty())
    {
      std::cout << "empty error of image file." << std::endl;
    }
    else
    {
      cv::cvtColor(img, img, CV_BGR2RGB);
      glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
      glGenTextures( 1, &_texname );
      glBindTexture( GL_TEXTURE_2D, _texname );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
      // glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, img.cols, img.rows, 0, GL_RGB, GL_UNSIGNED_BYTE, img.data );
      gluBuild2DMipmaps( GL_TEXTURE_2D, 3, img.cols, img.rows, GL_RGB, GL_UNSIGNED_BYTE, img.data );
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdMakeRenderObject
 *  Description     : This function sets the object data of OpenGL.
 *  Argument        : object data(z coordinate, texture coordinate)
 *  Return          : nothing special
 **************************************************************************************************/
void nvdMakeRenderObject( void )
{
  _displayList = glGenLists( 1 );
  glNewList( _displayList, GL_COMPILE );

  glBindTexture( GL_TEXTURE_2D, _texname );

  for( uword_type y = 0; y < (WINDOW_SIZE_H - 1); ++y )
  {
    for( uword_type x = 0; x < (WINDOW_SIZE_W -1); ++x )
    {
      // UV start
      glEnable( GL_TEXTURE_2D );

      glBegin( GL_POLYGON );
      //  (0, 0)
      glTexCoord2d( m_uv[y][x].u, m_uv[y][x].v );
      glVertex3f( m_vertex[y][x].x, m_vertex[y][x].y, (m_vertex[y][x].z * _render_size) );
      //  (0, 1)
      glTexCoord2d( m_uv[y][x+1].u, m_uv[y][x+1].v );
      glVertex3f( m_vertex[y][x+1].x, m_vertex[y][x+1].y, (m_vertex[y][x+1].z * _render_size) );
      //  (1, 1 )
      glTexCoord2d( m_uv[y+1][x+1].u, m_uv[y+1][x+1].v );
      glVertex3f( m_vertex[y+1][x+1].x, m_vertex[y+1][x+1].y, (m_vertex[y+1][x+1].z * _render_size) );
      //  (1, 0 )
      glTexCoord2d( m_uv[y+1][x].u, m_uv[y+1][x].v );
      glVertex3f( m_vertex[y+1][x].x, m_vertex[y+1][x].y, (m_vertex[y+1][x].z * _render_size) );

      glEnd();

      //  UV end
      glDisable( GL_TEXTURE_2D );
    }
  }

  glBindTexture( GL_TEXTURE_2D, 0 );

  glEndList();
}

/***************************************************************************************************
 *  Function        : nvdRotationRender
 *  Description     : This function rotates the polygon by 180 degrees.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void nvdRotationRender(void)
{
  glPushMatrix();
  glRotatef(180, 1.0, 0.0, 0.0);
  glTranslated( -(WINDOW_SIZE_W / 2.0), -(WINDOW_SIZE_H / 2.0), 0 );
  glCallList( _displayList );
  glPopMatrix();
}

/***************************************************************************************************
 *  Function        : bRead3DshapeFile
 *  Description     : This function reads an object file.
 *  Argument        : object file
 *  Return          : nothing special
 **************************************************************************************************/
bool_type bRead3DshapeFile( const char_type *filename )
{
  char_type buf[LENGTH_MAX];
  std::ifstream file;

  int x = WORK_DATA_INIT;
  int y = WORK_DATA_INIT;
  float_type depth_z = DEPTH_DATA_INIT;
  float_type depth_u = DEPTH_DATA_INIT;
  float_type depth_v = DEPTH_DATA_INIT;

  if( filename == NULL )
  {
    return FALSE;
  }

  file.open( filename, std::ios::in );
  if( !file.is_open() )
  {
    return FALSE;
  }

  while( !file.eof() )
  {
    file.getline( buf, sizeof( buf ) );

    switch( buf[0] )
    {
      case 'd':
        if( sscanf( (buf+2), "%d %d %f", &y, &x, &depth_z ) != 3)
        {
          return FALSE;
        }
        m_vertex[y][x].z = depth_z;

        break;

      case 'u':
        if( sscanf( (buf+2), "%d %d %f", &y, &x, &depth_u ) != 3)
        {
          return FALSE;
        }
        m_uv[y][x].u = depth_u;

        break;

      case 'v':
        if( sscanf( (buf+2), "%d %d %f", &y, &x, &depth_v ) != 3)
        {
          return FALSE;
        }
        m_uv[y][x].v = depth_v;

        break;

      default:
        break;
    }
  }

  return TRUE;
}

/***************************************************************************************************
 *  Function        : nvdObjectDataInitialize
 *  Description     : This function initializes object data.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void nvdObjectDataInitialize(void)
{
  uword_type width  = WINDOW_SIZE_W;
  uword_type height = WINDOW_SIZE_H;

  for( uword_type y = 0; y < height; ++y ){
    for( uword_type x = 0; x < width; ++x ){
      m_vertex[y][x].y = static_cast<float_type>(y);
      m_vertex[y][x].x = static_cast<float_type>(x);
      m_vertex[y][x].z = Z_DATA_INIT;

      // UV ( 0.0f ~ 1.0f )
      m_uv[y][x].v = y * ( V_DATA_INIT / (height-1) );
      m_uv[y][x].u = x * ( U_DATA_INIT / (width-1) );
    }
  }

  _render_size = RENDER_SIZE_INIT;
}

/***************************************************************************************************
 *  Function        : nvdCheckOpenGLRequest
 *  Description     : This function checks the request to OpenGL display.
 *  Argument        : object file
 *  Return          : nothing special
 **************************************************************************************************/
void nvdCheckOpenGLRequest( void )
{
  if( bReqFileChange )
  {
    nvdOpenGLFileCopy();
  }

  if( bReqWireFrame )
  {
    nvdOpenGLChangeWireFrame();
  }

  if( bReqZcoordAdjust )
  {
    nvdOpenGLChangeZcoordinate();
  }

  if( bReqPlayMode )
  {
    nvdOpenGLChangePlayMode();
  }
}

/***************************************************************************************************
 *  Function        : nvdOpenGLFileCopy
 *  Description     : This function copies the OpenGL object file.
 *  Argument        : object file
 *  Return          : nothing special
 **************************************************************************************************/
void nvdOpenGLFileCopy( void )
{
  bReqFileChange = FALSE;
  bVaidDiapFlag  = FALSE;

  if( PathFileExists( src_path_ftv ) && PathFileExists( src_path_bmp ) )
  {
    if( PathFileExists( dst_path_ftv ) )
    {
      (void)DeleteFile( dst_path_ftv );
    }
    if( PathFileExists( dst_path_bmp ) )
    {
      (void)DeleteFile( dst_path_bmp );
    }

    if( CopyFile(src_path_ftv, dst_path_ftv, FALSE) && CopyFile(src_path_bmp, dst_path_bmp, FALSE) )
    {
      if( bRead3DshapeFile(&dst_path_ftv[0]) )
      {
        std::cout << READ_FILE_SUCCESS << std::endl;
        bVaidDiapFlag = TRUE;
      }
      else
      {
        nvdObjectDataInitialize();
      }
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdOpenGLChangeWireFrame
 *  Description     : This function sets the wireframe mode.
 *  Argument        : wireframe mode
 *  Return          : nothing special
 **************************************************************************************************/
void nvdOpenGLChangeWireFrame( void )
{
  bReqWireFrame = FALSE;

  if(uwWireFrameMode == DEF_WIRE_FRAME_ON)
  {
    bWireFrameMode = DEF_WIRE_FRAME_ON;
  }
  else
  {
    bWireFrameMode = DEF_WIRE_FRAME_OFF;
  }
}

/***************************************************************************************************
 *  Function        : nvdOpenGLChangeZcoordinate
 *  Description     : This function sets the Z coordinate value.
 *  Argument        : Z coordinate value
 *  Return          : nothing special
 **************************************************************************************************/
void nvdOpenGLChangeZcoordinate( void )
{
  bReqZcoordAdjust = FALSE;

  if( (g_fZcoordValue >= RENDER_SIZE_INIT) && (g_fZcoordValue < RENDER_SIZE_MAX) )
  {
    _render_size = g_fZcoordValue;
  }
}

/***************************************************************************************************
 *  Function        : nvdOpenGLChangePlayMode
 *  Description     : This function sets the play or stop mode.
 *  Argument        : play or stop mode
 *  Return          : nothing special
 **************************************************************************************************/
void nvdOpenGLChangePlayMode( void )
{
  bReqPlayMode = FALSE;

  if(bPlayMode_sts == DEF_PLAY_MODE)
  {
    bPlayStopMode = DEF_PLAY_MODE;
  }
  else
  {
    bPlayStopMode = DEF_STOP_MODE;
  }
}

/***************************************************************************************************
 *  Function        : nvd3DImageViewerOnInitialize
 *  Description     : When onInitialize, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvd3DImageViewerOnInitialize( void )
{
  std::cout << _ON_INITIALIZE << std::endl;

  DWORD result = 0;

  char_type strPath[DEF_MAX_LENGTH];
  char_type strDrive[DEF_MIN_LENGTH];
  char_type strDir[DEF_MAX_LENGTH];
  char_type strFile[DEF_MID_LENGTH];
  char_type strExt[DEF_MIN_LENGTH];

  memset(strPath, 0x00, sizeof(strPath));
  memset(strDrive, 0x00, sizeof(strDrive));
  memset(strDir, 0x00, sizeof(strDir));
  memset(strFile, 0x00, sizeof(strFile));
  memset(strExt, 0x00, sizeof(strExt));
  memset(g_absolute_dir, 0x00, sizeof(g_absolute_dir));

  uwAbsolute_dir_err = DEF_GET_PATH_NG;
  result = GetModuleFileName(NULL, strPath, sizeof(strPath));
  if( result != DEF_GET_PATH_NG )
  {
    uwAbsolute_dir_err = DEF_GET_PATH_OK;
    _splitpath(strPath, strDrive, strDir, strFile, strExt);
    sprintf(g_absolute_dir, "%s%s%s", strDrive, strDir, "3d_shape");
    std::cout << "absolute dir path: " << g_absolute_dir << std::endl;
  }
}

/***************************************************************************************************
 *  Function        : b3DImageViewerOnActivated
 *  Description     : When onActivated, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_3DImageShapeViewer::b3DImageViewerOnActivated( void )
{
  bool_type abReturn = FALSE;

  std::cout << _ON_ACTIVATED << std::endl;

  bReqFileChange = FALSE;
  bVaidDiapFlag  = FALSE;

  bReqWireFrame   = FALSE;
  bWireFrameMode  = DEF_WIRE_FRAME_OFF;
  uwWireFrameMode = DEF_WIRE_FRAME_OFF;

  bReqZcoordAdjust = FALSE;
  g_fZcoordValue   = RENDER_SIZE_INIT;

  bReqPlayMode  = FALSE;
  bPlayMode_sts = DEF_STOP_MODE;
  bPlayStopMode = DEF_STOP_MODE;

  g_RGB_Red   = fColorCodeConversion(m_Back_Color_Red);
  g_RGB_Green = fColorCodeConversion(m_Back_Color_Green);
  g_RGB_Blue  = fColorCodeConversion(m_Back_Color_Blue);

  winname = "Check images";
  createAppWindow();

  bEXEC_Thread = EXEC_INIT;

  if( uwAbsolute_dir_err == DEF_GET_PATH_OK )
  {
    if( bAllocationDynamicmMem(WINDOW_SIZE_W, WINDOW_SIZE_H) )
    {
      if( bThreadStart() )
      {
        abReturn = TRUE;
      }
    }
  }

  return abReturn;
}

/***************************************************************************************************
*  Function        : fColorCodeConversion
*  Description     : This function converts the color code.
*  Argument        : RGB Color code (Normal type: 0-255)
*  Return          : RGB Color code (OpenGL type: 0.0-1.0)
**************************************************************************************************/
float_type rtc_3DImageShapeViewer::fColorCodeConversion( sword_type RGB255)
{
  float_type fRGB01 = (float_type)RGB255 / DEF_RGB_255F;

  if (fRGB01 < DEF_RGB_GL0F)
  {
    fRGB01 = DEF_RGB_GL0F;
  }
  else if (fRGB01 > DEF_RGB_GL1F)
  {
    fRGB01 = DEF_RGB_GL1F;
  }

  return fRGB01;
}

/***************************************************************************************************
 *  Function        : bAllocationDynamicmMem
 *  Description     : This function allocates dynamic memory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
bool_type rtc_3DImageShapeViewer::bAllocationDynamicmMem( uword_type width, uword_type height )
{
  try{
    m_vertex = new Vertex*[ height ];
    m_uv     = new UV*[ height ];

    for( uword_type i = 0; i < height; i++ ){
      m_vertex[i] = new Vertex[ width ];
      m_uv[i]     = new UV[ width ];
    }
  }catch( std::bad_alloc ){
    return FALSE;
  }

  for( uword_type y = 0; y < height; ++y ){
    for( uword_type x = 0; x < width; ++x ){
      m_vertex[y][x].y = static_cast<float_type>(y);
      m_vertex[y][x].x = static_cast<float_type>(x);
      m_vertex[y][x].z = Z_DATA_INIT;

      // UV ( 0.0f ~ 1.0f )
      m_uv[y][x].v = y * ( V_DATA_INIT / (height-1) );
      m_uv[y][x].u = x * ( U_DATA_INIT / (width-1) );
    }
  }

  _render_size = RENDER_SIZE_INIT;

  return TRUE;
}

/***************************************************************************************************
 *  Function        : nvdRelease
 *  Description     : 
 *  Argument        : 
 *  Return          : 
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdRelease( void )
{
  uword_type height = WINDOW_SIZE_H;

  if( m_vertex != NULL && m_uv != NULL )
  {
    for( uword_type i = 0; i < height; ++i )
    {
      delete [] m_vertex[i];
      delete [] m_uv[i];

      m_vertex[i] = NULL;
      m_uv[i] = NULL;
    }
    delete [] m_uv;
    delete [] m_vertex;

    m_uv = NULL;
    m_vertex = NULL;
  }
}

/***************************************************************************************************
 *  Function        : nvd3DImageViewerOnDeactivated
 *  Description     : When onDeactivated, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvd3DImageViewerOnDeactivated( void )
{
  closeAppWindow();

  sword_type ret = glutGetWindow();
  if( ret )
  {
    glutDestroyWindow( glutGetWindow() );
    std::cout << _DELETE_OPENGL_WINDOW << std::endl;

    uword_type uwLoopCnt = GUARD_TIM_INI;
    bool_type  bLoopFlag = TRUE;

    while( bLoopFlag == TRUE )
    {
      Sleep(OPENGL_CLOSE);
      if( bEXEC_Thread == EXEC_END )
      {
        bLoopFlag = FALSE;
        Sleep(THREAD_CLOSE);
      }
      else
      {
        uwLoopCnt = uwLoopCnt + 1;
        if( uwLoopCnt > GUARD_TIM_MAX )
        {
          bLoopFlag = FALSE;
        }
      }
    }

  }

  WaitForSingleObject( hHandle, INFINITE );
  if( hHandle != NULL )
  {
    CloseHandle ( hHandle );
    std::cout << _DELETE_OPENGL_THREAD << std::endl;
  }

  nvdRelease();
  g_ViewportCamera.Reset();

  std::cout << _ON_DEACTIVATED << std::endl;
}

/***************************************************************************************************
 *  Function        : nvd3DImageViewerOnAborting
 *  Description     : When onAborting, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvd3DImageViewerOnAborting( void )
{
  std::cout << _ON_ABORTING << std::endl;
  std::cout << _ERROR_RESET << std::endl;
}

/***************************************************************************************************
 *  Function        : nvd3DImageViewerOnReset
 *  Description     : When onReset, this function execute the initialization.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvd3DImageViewerOnReset( void )
{
  std::cout << _ON_RESET << std::endl;
}

/***************************************************************************************************
 *  Function        : nvd3DImageViewerOnExecute
 *  Description     : This function calls a function that checks the input port.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvd3DImageViewerOnExecute( void )
{
  nvdCheckInputDirectoryPath();
  nvdCheckInputWireFrameMode();
  nvdCheckInputZcoordinate();
  nvdCheckInputPlayStopMode();
  readCameraImage();
}

/***************************************************************************************************
 *  Function        : nvdCheckInputDirectoryPath
 *  Description     : This function checks the input directory path. Then copy the object file.
 *  Argument        : directory path
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdCheckInputDirectoryPath( void )
{
  std::string file_original = DF_EMPTY_STRING;
  std::string file_path_ftv = DF_EMPTY_STRING;
  std::string file_path_bmp = DF_EMPTY_STRING;

  char_type  char_path_org[DEF_MAX_LENGTH];
  char_type  char_path_ftv[DEF_MAX_LENGTH];
  char_type  char_path_bmp[DEF_MAX_LENGTH];

  memset(char_path_org, 0x00, sizeof(char_path_org));
  memset(char_path_ftv, 0x00, sizeof(char_path_ftv));
  memset(char_path_bmp, 0x00, sizeof(char_path_bmp));

  if( m_file_pathIn.isNew() )
  {
    m_file_pathIn.read();
    file_original = m_file_path.data;

    strcpy( char_path_org, file_original.c_str() );
    if( PathIsDirectory(char_path_org) )
    {
      file_path_ftv = file_original + "/3d_shape.ftv";
      strcpy( char_path_ftv, file_path_ftv.c_str() );

      file_path_bmp = file_original + "/3d_shape.bmp";
      strcpy( char_path_bmp, file_path_bmp.c_str() );

      if( PathFileExists( char_path_ftv ) && PathFileExists( char_path_bmp ) )
      {
        nvdCreationDir3d_shape();
        nvdDeleteObjFile();
        nvdCopyObjFile( &char_path_ftv[0], &char_path_bmp[0] );
      }
    }
    else
    {
      std::cout << _ERROR_PATH_WRONG << std::endl;
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdCheckInputWireFrameMode
 *  Description     : This function checks the input of wireframe mode.
 *  Argument        : wireframe mode
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdCheckInputWireFrameMode( void )
{
  std::string wire_frame_mode = DF_EMPTY_STRING;

  if( m_wire_frame_modeIn.isNew() )
  {
    m_wire_frame_modeIn.read();
    wire_frame_mode = m_wire_frame_mode.data;
    if( wire_frame_mode == DEFINE_FRAME_ON)
    {
      bReqWireFrame   = TRUE;
      uwWireFrameMode = DEF_WIRE_FRAME_ON;
    }
    else if(wire_frame_mode == DEFINE_FRAME_OFF)
    {
      bReqWireFrame   = TRUE;
      uwWireFrameMode = DEF_WIRE_FRAME_OFF;
    }
    else
    {
      std::cout << _ERROR_INPUT_DATA << std::endl;
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdCheckInputZcoordinate
 *  Description     : This function checks input of Z coordinate adjustment value.
 *  Argument        : Z coordinate adjustment value
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdCheckInputZcoordinate( void )
{
  std::string sz_coord_wk1 = DF_EMPTY_STRING;
  float_type  fz_coord_wk2 = Z_DATA_INIT;
  float_type  fz_coord_val = Z_DATA_INIT;

  bool_type bResult = TRUE;

  if( m_z_adjustIn.isNew() )
  {
    m_z_adjustIn.read();
    sz_coord_wk1 = m_z_adjust.data;

    try{
      fz_coord_wk2 = static_cast<float_type>(std::stoi(sz_coord_wk1));
    }
    catch (...){
      bResult = FALSE;
    }

    if( bResult )
    {
      if( (fz_coord_wk2 >= Z_COORD_MIN) && (fz_coord_wk2 <= Z_COORD_MAX) )
      {
        bReqZcoordAdjust = TRUE;
        fz_coord_val     = static_cast<float_type>((fz_coord_wk2/10));
        g_fZcoordValue   = Z_COORD_VAL + fz_coord_val;
      }
      else
      {
        std::cout << _ERROR_INPUT_DATA << std::endl;
      }
    }
    else
    {
      std::cout << _ERROR_INPUT_DATA << std::endl;
    }
  }
}

/***************************************************************************************************
 *  Function        : nvdCheckInputPlayStopMode
 *  Description     : This function checks the input of play or stop mode.
 *  Argument        : play or stop mode
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdCheckInputPlayStopMode( void )
{
  std::string ply_mode = DF_EMPTY_STRING;

  if( m_ply_modeIn.isNew() )
  {
    m_ply_modeIn.read();
    ply_mode = m_ply_mode.data;
    if( ply_mode == DEFINE_MODE_PLY)
    {
      bReqPlayMode  = TRUE;
      bPlayMode_sts = DEF_PLAY_MODE;
    }
    else if(ply_mode == DEFINE_MODE_STP)
    {
      bReqPlayMode  = TRUE;
      bPlayMode_sts = DEF_STOP_MODE;
    }
    else
    {
      std::cout << _ERROR_INPUT_DATA << std::endl;
    }
  }
}

/***************************************************************************************************
 *  Function        : createAppWindow
 *  Description     : This function creates an initial window.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::createAppWindow( void )
{
  std::cout << _CREATE_WINDOW << std::endl;

  cv::namedWindow(winname, cv::WINDOW_NORMAL);
  cv::resizeWindow(winname, WINDOW_SIZE_W, WINDOW_SIZE_H);
}

/***************************************************************************************************
 *  Function        : closeAppWindow
 *  Description     : This function closes the all window.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::closeAppWindow( void )
{
  std::cout << _DELETE_WINDOW << std::endl;
  cv::destroyAllWindows();
}

/***************************************************************************************************
 *  Function        : readCameraImage
 *  Description     : This function reads camera-image type data.
 *  Argument        : camera-image type data
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::readCameraImage( void )
{
  if( m_in_imgIn.isNew() )
  {
    m_in_imgIn.read();
    showCameraImage( m_in_img );
  }
  cv::waitKey(10);
}

/***************************************************************************************************
 *  Function        : closeAppWindow
 *  Description     : This function checks the size of the image.
 *                  : Also, when there is no problem with the image size, the image is displayed.
 *  Argument        : camera-image type data
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::showCameraImage( RTC::CameraImage image )
{
  bool_type  retVal = FALSE;
  uword_type imglen = 0;

  if( (image.width == IMAGE_SIZE_W_MIN) || (image.height == IMAGE_SIZE_H_MIN) )
  {
    std::cout << _ERROR_IMAGE_EMPTY << std::endl;
  }
  else if( (image.width <= IMAGE_SIZE_W_MAX) && (image.height <= IMAGE_SIZE_H_MAX) )
  {
    retVal = TRUE;
  }
  else
  {
    std::cout << _ERROR_IMAGE_SIZE << std::endl;
  }

  if( retVal )
  {
    imglen = image.pixels.length();
    cv::Mat imagebuf = cv::Mat( cv::Size(image.width, image.height), CV_8UC3 );
    memcpy(imagebuf.data, (void *)&(image.pixels[0]), imglen);
    cv::namedWindow( winname, cv::WINDOW_NORMAL );
    cv::imshow( winname, imagebuf );
  }
}

/***************************************************************************************************
 *  Function        : nvdCreationDir3d_shape
 *  Description     : This function creates a 3d-shape directory when there is no 3d-shape directory.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdCreationDir3d_shape( void )
{
  if( !PathIsDirectory(g_absolute_dir) )
  {
    (void)_mkdir("3d_shape");
  }
}

/***************************************************************************************************
 *  Function        : nvdDeleteObjFile
 *  Description     : This function deletes the object file if an object file exists.
 *  Argument        : nothing special
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdDeleteObjFile( void )
{
  if( PathFileExists( "./3d_shape/3d_shape.ftv" ) )
  {
    (void)DeleteFile("./3d_shape/3d_shape.ftv");
  }
  if( PathFileExists( "./3d_shape/3d_shape.bmp" ) )
  {
    (void)DeleteFile("./3d_shape/3d_shape.bmp");
  }
}

/***************************************************************************************************
 *  Function        : nvdCopyObjFile
 *  Description     : This function copies the object file. Then set the request flag of image update.
 *  Argument        : source file path
 *  Return          : nothing special
 **************************************************************************************************/
void rtc_3DImageShapeViewer::nvdCopyObjFile( char_type *sftv_buf, char_type *sbmp_buf )
{
  char_type  dftv_buf[DEF_MAX_LENGTH];
  char_type  dbmp_buf[DEF_MAX_LENGTH];
  char_type  tftv_buf[DEF_MAX_LENGTH];
  char_type  tbmp_buf[DEF_MAX_LENGTH];

  memset(dftv_buf, 0x00, sizeof(dftv_buf));
  memset(dbmp_buf, 0x00, sizeof(dbmp_buf));
  memset(tftv_buf, 0x00, sizeof(tftv_buf));
  memset(tbmp_buf, 0x00, sizeof(tbmp_buf));

  sprintf(dftv_buf, "%s%s", g_absolute_dir, "\\3d_shape.ftv");
  sprintf(dbmp_buf, "%s%s", g_absolute_dir, "\\3d_shape.bmp");
  sprintf(tftv_buf, "%s%s", g_absolute_dir, "\\3d_shape_opngl.ftv");
  sprintf(tbmp_buf, "%s%s", g_absolute_dir, "\\3d_shape_opngl.bmp");

  if ( CopyFile(sftv_buf, dftv_buf, FALSE) && CopyFile(sbmp_buf, dbmp_buf, FALSE) )
  {
    strcpy(src_path_ftv, dftv_buf);
    strcpy(src_path_bmp, dbmp_buf);
    strcpy(dst_path_ftv, tftv_buf);
    strcpy(dst_path_bmp, tbmp_buf);

    bReqFileChange = TRUE;
  }
}

/***************************************************************************************************
 **************************************************************************************************/


extern "C"
{
 
  void rtc_3DImageShapeViewerInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_3dimageshapeviewer_spec);
    manager->registerFactory(profile,
                             RTC::Create<rtc_3DImageShapeViewer>,
                             RTC::Delete<rtc_3DImageShapeViewer>);
  }
  
};

